@@ -34,6 +34,10 @@ _add('-n', '--site-name', dest='site_name',
     help='Set the name of the site to NAME', metavar='NAME')
 _add('-m', '--front-page', dest='front_page',
     help='Use PAGE as the front page', metavar='PAGE')
+_add('-a', '--alias-page', dest='alias_page',
+    help='Use PAGE as the alias page', metavar='PAGE')
+_add('-H', '--help-page', dest='help_page',
+    help='Use PAGE as the editor help text', metavar='PAGE')
 _add('-e', '--encoding', dest='page_charset',
     help='Use encoding ENC to read and write pages', metavar='ENC')
 _add('-c', '--config-file', dest='config_file',
@@ -50,6 +54,9 @@ _add('-W', '--wiki-words', dest='wiki_words',
     help='Enable WikiWord links', action="store_true")
 _add('-I', '--ignore-indent', dest='ignore_indent',
     help='Treat indented lines as normal text', action="store_true")
+_add('-M', '--math-url', dest='math_url',
+    help='Use the URL for rendering the equations. Empty to disable math.',
+    metavar='URL')
 _add('-P', '--pygments-style', dest='pygments_style',
     help='Use the STYLE pygments style for highlighting',
     metavar='STYLE')
