@@ -31,8 +31,55 @@ try:
 except ImportError:
     pass
 
-from hatta import error
-from hatta import parser
+import hatta.error
+import hatta.parser
+
+
+def check_lock(wiki, title):
+    _ = wiki.gettext
+    restricted_pages = [
+        'scripts.js',
+        'robots.txt',
+    ]
+    if wiki.read_only:
+        raise hatta.error.ForbiddenErr(_(u"This site is read-only."))
+    if title in restricted_pages:
+        raise hatta.error.ForbiddenErr(_(u"""Can't edit this page.
+It can only be edited by the site admin directly on the disk."""))
+    if title in wiki.index.page_links(wiki.locked_page):
+        raise hatta.error.ForbiddenErr(_(u"This page is locked."))
+
+
+
+def get_page(request, title):
+    """Creates a page object based on page's mime type"""
+
+    if title:
+        try:
+            page_class, mime = request.wiki.filename_map[title]
+        except KeyError:
+            mime = page_mime(title)
+            major, minor = mime.split('/', 1)
+            try:
+                page_class = request.wiki.mime_map[mime]
+            except KeyError:
+                try:
+                    plus_pos = minor.find('+')
+                    if plus_pos > 0:
+                        minor_base = minor[plus_pos:]
+                    else:
+                        minor_base = ''
+                    base_mime = '/'.join([major, minor_base])
+                    page_class = request.wiki.mime_map[base_mime]
+                except KeyError:
+                    try:
+                        page_class = request.wiki.mime_map[major]
+                    except KeyError:
+                        page_class = request.wiki.mime_map['']
+    else:
+        page_class = WikiPageSpecial
+        mime = ''
+    return page_class(request.wiki, request, title, mime)
 
 
 def page_mime(title):
@@ -121,7 +168,7 @@ class WikiPage(object):
             classes = [class_]
         else:
             classes = []
-        if parser.external_link(addr):
+        if hatta.parser.external_link(addr):
             classes.append('external')
             if addr.startswith('mailto:'):
                 # Obfuscate e-mails a little bit.
@@ -164,7 +211,7 @@ class WikiPage(object):
         addr = addr.strip()
         html = werkzeug.html
         chunk = ''
-        if parser.external_link(addr):
+        if hatta.parser.external_link(addr):
             return html.img(src=werkzeug.url_fix(addr), class_="external",
                             alt=alt)
         if '#' in addr:
@@ -209,9 +256,9 @@ class WikiPage(object):
         edit_url = None
         if self.title:
             try:
-                self.wiki._check_lock(self.title)
-                edit_url = self.get_url(self.title, self.wiki.edit)
-            except error.ForbiddenErr:
+                check_lock(self.wiki, self.title)
+                edit_url = self.get_url(self.title, 'edit')
+            except hatta.error.ForbiddenErr:
                 pass
         context = {
             'request': self.request,
@@ -238,11 +285,18 @@ class WikiPage(object):
                 dependencies.add(werkzeug.url_quote(title))
         for title in [self.wiki.menu_page]:
             if title in self.storage:
-                inode, size, mtime = self.storage.page_file_meta(title)
-                etag = '%s/%d-%d' % (werkzeug.url_quote(title), inode, mtime)
+                rev, date, author, comment = self.storage.page_meta(title)
+                etag = '%s/%d-%s' % (werkzeug.url_quote(title), rev, date.isoformat())
                 dependencies.add(etag)
         return dependencies
 
+    def get_edit_help(self):
+        page = get_page(self.request, self.wiki.help_page)
+        try:
+            return ''.join(page.view_content())
+        except hatta.error.NotFoundErr:
+            return ''
+
     def render_editor(self, preview=None, captcha_error=None):
         _ = self.wiki.gettext
         author = self.request.get_author()
@@ -260,9 +314,14 @@ class WikiPage(object):
                     self.wiki.recaptcha_public_key, error=captcha_error)
         else:
             recaptcha_html = None
-        return self.template('edit_file.html', comment=comment,
-                             recaptcha_html=recaptcha_html,
-                             author=author, parent=rev)
+        context = {
+            'comment': comment,
+            'author': author,
+            'parent': rev,
+            'recaptcha_html': recaptcha_html,
+            'help': self.get_edit_help(),
+        }
+        return self.template('edit_file.html', **context)
 
 
 class WikiPageSpecial(WikiPage):
@@ -292,8 +351,7 @@ class WikiPageText(WikiPage):
         """
 
         if lines is None:
-            f = self.storage.open_page(self.title)
-            lines = self.storage.page_lines(f)
+            lines = self.storage.page_text(self.title).splitlines(True)
         return self.content_iter(lines)
 
     def render_editor(self, preview=None, captcha_error=None):
@@ -303,17 +361,16 @@ class WikiPageText(WikiPage):
         author = self.request.get_author()
         lines = []
         try:
-            page_file = self.storage.open_page(self.title)
-            lines = self.storage.page_lines(page_file)
+            lines = self.storage.page_text(self.title).splitlines(True)
             (rev, old_date, old_author,
                 old_comment) = self.storage.page_meta(self.title)
             comment = _(u'modified')
             if old_author == author:
                 comment = old_comment
-        except error.NotFoundErr:
+        except hatta.error.NotFoundErr:
             comment = _(u'created')
             rev = -1
-        except error.ForbiddenErr, e:
+        except hatta.error.ForbiddenErr, e:
             return werkzeug.html.p(werkzeug.html(unicode(e)))
         if preview:
             lines = preview
@@ -323,10 +380,16 @@ class WikiPageText(WikiPage):
                     self.wiki.recaptcha_public_key, error=captcha_error)
         else:
             recaptcha_html = None
-        return self.template('edit_text.html', comment=comment,
-                             preview=preview,
-                             recaptcha_html=recaptcha_html,
-                             author=author, parent=rev, lines=lines)
+        context = {
+            'comment': comment,
+            'preview': preview,
+            'recaptcha_html': recaptcha_html,
+            'help': self.get_edit_help(),
+            'author': author,
+            'parent': rev,
+            'lines': lines,
+        }
+        return self.template('edit_text.html', **context)
 
     def diff_content(self, from_text, to_text, message=u''):
         """Generate the HTML markup for a diff."""
@@ -434,9 +497,9 @@ class WikiPageWiki(WikiPageColorText):
     def __init__(self, *args, **kw):
         super(WikiPageWiki, self).__init__(*args, **kw)
         if self.config.get_bool('wiki_words', False):
-            self.parser = parser.WikiWikiParser
+            self.parser = hatta.parser.WikiWikiParser
         else:
-            self.parser = parser.WikiParser
+            self.parser = hatta.parser.WikiParser
         if self.config.get_bool('ignore_indent', False):
             try:
                 del self.parser.block['indent']
@@ -449,14 +512,13 @@ class WikiPageWiki(WikiPageColorText):
         if text is None:
             try:
                 text = self.storage.page_text(self.title)
-            except error.NotFoundErr:
+            except hatta.error.NotFoundErr:
                 text = u''
         return self.parser.extract_links(text)
 
     def view_content(self, lines=None):
         if lines is None:
-            f = self.storage.open_page(self.title)
-            lines = self.storage.page_lines(f)
+            lines = self.storage.page_text(self.title).splitlines(True)
         if self.wiki.icon_page and self.wiki.icon_page in self.storage:
             icons = self.index.page_links_and_labels(self.wiki.icon_page)
             smilies = dict((emo, link) for (link, emo) in icons)
@@ -466,22 +528,28 @@ class WikiPageWiki(WikiPageColorText):
                              self.highlight, self.wiki_math, smilies)
         return content
 
-    def wiki_math(self, math):
-        math_url = self.config.get('math_url',
-                            'http://www.mathtran.org/cgi-bin/mathtran?tex=')
+    def wiki_math(self, math_text, display=False):
+        math_url = self.wiki.math_url
+        if math_url == '':
+            return werkzeug.escape(math_text)
+        elif math_url == 'mathjax':
+            if display:
+                return werkzeug.escape(u"$$\n%s\n$$" % math_text)
+            else:
+                return werkzeug.escape(u"$%s$" % math_text)
         if '%s' in math_url:
-            url = math_url % werkzeug.url_quote(math)
+            url = math_url % werkzeug.url_quote(math_text)
         else:
-            url = '%s%s' % (math_url, werkzeug.url_quote(math))
-        label = werkzeug.escape(math, quote=True)
+            url = '%s%s' % (math_url, werkzeug.url_quote(math_text))
+        label = werkzeug.escape(math_text, quote=True)
         return werkzeug.html.img(src=url, alt=label, class_="math")
 
     def dependencies(self):
         dependencies = WikiPage.dependencies(self)
         for title in [self.wiki.icon_page, self.wiki.alias_page]:
             if title in self.storage:
-                inode, size, mtime = self.storage.page_file_meta(title)
-                etag = '%s/%d-%d' % (werkzeug.url_quote(title), inode, mtime)
+                rev, date, author, comment = self.storage.page_meta(title)
+                etag = '%s/%d-%s' % (werkzeug.url_quote(title), rev, date.isoformat())
                 dependencies.add(etag)
         for link in self.index.page_links(self.title):
             if link not in self.storage:
@@ -494,7 +562,7 @@ class WikiPageFile(WikiPage):
 
     def view_content(self, lines=None):
         if self.title not in self.storage:
-            raise error.NotFoundErr()
+            raise hatta.error.NotFoundErr()
         content = ['<p>Download <a href="%s">%s</a> as <i>%s</i>.</p>' %
                    (self.request.get_download_url(self.title),
                     werkzeug.escape(self.title), self.mime)]
@@ -508,9 +576,9 @@ class WikiPageImage(WikiPageFile):
 
     def view_content(self, lines=None):
         if self.title not in self.storage:
-            raise error.NotFoundErr()
+            raise hatta.error.NotFoundErr()
         content = ['<img src="%s" alt="%s">'
-                   % (self.request.get_url(self.title, self.wiki.render),
+                   % (self.request.get_url(self.title, 'render'),
                       werkzeug.escape(self.title))]
         return content
 
@@ -535,7 +603,7 @@ class WikiPageImage(WikiPageFile):
             im.thumbnail((128, 128), Image.ANTIALIAS)
             im.save(cache_file, 'PNG')
         except IOError:
-            raise error.UnsupportedMediaTypeErr('Image corrupted')
+            raise hatta.error.UnsupportedMediaTypeErr('Image corrupted')
         cache_file.close()
         return cache_path
 
@@ -567,7 +635,7 @@ class WikiPageCSV(WikiPageFile):
 
     def view_content(self, lines=None):
         if self.title not in self.storage:
-            raise error.NotFoundErr()
+            raise hatta.error.NotFoundErr()
         return self.content_iter(lines)
 
 
