@@ -2,16 +2,14 @@
 # -*- coding: utf-8 -*-
 
 import datetime
-import errno
 import os
 import re
-import tempfile
 import thread
 import werkzeug
+import StringIO
 
 # Note: we have to set these before importing Mercurial
 os.environ['HGENCODING'] = 'utf-8'
-os.environ['HGMERGE'] = "internal:merge"
 
 import mercurial.commands
 import mercurial.hg
@@ -20,6 +18,7 @@ import mercurial.merge
 import mercurial.revlog
 import mercurial.ui
 import mercurial.util
+import mercurial.simplemerge
 
 from hatta import error
 from hatta import page
@@ -58,6 +57,19 @@ def _get_ui():
     return ui
 
 
+def merge_func(base, other, this):
+    """Used for merging edit conflicts."""
+
+    if (mercurial.util.binary(this) or mercurial.util.binary(base) or
+        mercurial.util.binary(other)):
+        raise ValueError("can't merge binary data")
+    m3 = mercurial.simplemerge.Merge3Text(base, this, other)
+    return ''.join(m3.merge_lines(start_marker='<<<<<<< local',
+                                  mid_marker='=======',
+                                  end_marker='>>>>>>> other',
+                                  base_marker=None))
+
+
 def _get_datetime(filectx):
     """
     Create a datetime object for the changeset's cretion time, taking into
@@ -105,11 +117,13 @@ class WikiStorage(object):
             # Create the repository if needed.
             mercurial.hg.repository(self.ui, self.repo_path, create=True)
         self._repos = {}
+        self._tips = {}
 
     def reopen(self):
         """Close and reopen the repo, to make sure we are up to date."""
 
         self._repos = {}
+        self._tips = {}
 
     @property
     def repo(self):
@@ -123,23 +137,6 @@ class WikiStorage(object):
             self._repos[thread_id] = repo
             return repo
 
-    def _check_path(self, path):
-        """
-        Ensure that the path is within allowed bounds.
-        """
-
-        _ = self._
-        abspath = os.path.abspath(path)
-        if os.path.islink(path) or os.path.isdir(path):
-            raise error.ForbiddenErr(
-                _(u"Can't use symbolic links or directories as pages"))
-        if not abspath.startswith(self.path):
-            raise error.ForbiddenErr(
-                _(u"Can't read or write outside of the pages repository"))
-
-    def _file_path(self, title):
-        return os.path.join(self.repo_path, self._title_to_file(title))
-
     def _title_to_file(self, title):
         title = unicode(title).strip()
         filename = werkzeug.url_quote(title, safe='')
@@ -168,120 +165,68 @@ class WikiStorage(object):
         return werkzeug.url_unquote(name)
 
     def __contains__(self, title):
-        if title:
-            file_path = self._file_path(title)
-            return os.path.isfile(file_path) and not os.path.islink(file_path)
+        repo_file = self._title_to_file(title)
+        return repo_file in self._changectx()
 
     def __iter__(self):
         return self.all_pages()
 
-    def merge_changes(self, changectx, repo_file, text, user, parent):
-        """Commits and merges conflicting changes in the repository."""
-
-        _ = self._
-        tip_node = changectx.node()
-        filectx = changectx[repo_file].filectx(parent)
-        parent_node = filectx.changectx().node()
-
-        self.repo.dirstate.setparents(parent_node)
-        node = self._commit([repo_file], text, user)
-
-        partial = lambda filename: repo_file == filename
+    def _get_parents(self, filename, parent_rev):
+        if parent_rev is None:
+            return 'tip', None
         try:
-            mercurial.merge.update(self.repo, tip_node, True, True, partial)
-            msg = _(u'merge of edit conflict')
-        except mercurial.util.Abort:
-            msg = _(u'failed merge of edit conflict')
-        self.repo.dirstate.setparents(tip_node, node)
-        # Mercurial 1.1 and later need updating the merge state
-        try:
-            mergestate = mercurial.merge.mergestate
-        except AttributeError:
-            pass
-        else:
-            state = mergestate(self.repo)
-            try:
-                state.mark(repo_file, "r")
-            except KeyError:
-                # There were no conflicts to mark
-                pass
-            else:
-                # Mercurial 1.7+ needs a commit
-                try:
-                    commit = state.commit
-                except AttributeError:
-                    pass
-                else:
-                    commit()
-        return msg
-
-    @locked_repo
-    def save_file(self, title, file_name, author=u'', comment=u'',
-                  parent=None):
-        """Save an existing file as specified page."""
-
-        _ = self._
-        user = author.encode('utf-8') or _(u'anon').encode('utf-8')
-        text = comment.encode('utf-8') or _(u'comment').encode('utf-8')
-        repo_file = self._title_to_file(title)
-        file_path = self._file_path(title)
-        self._check_path(file_path)
-        try:
-            mercurial.util.rename(file_name, file_path)
-        except OSError, e:
-            if e.errno == errno.ENAMETOOLONG:
-                # "File name too long"
-                raise error.RequestURITooLarge()
-            else:
-                raise
-        changectx = self._changectx()
-        try:
-            # Mercurial 1.5 and earlier have .add() on the repo
-            add = self.repo.add
-        except AttributeError:
-            # Mercurial 1.6
-            add = self.repo[None].add
-        try:
-            filectx_tip = changectx[repo_file]
-            current_page_rev = filectx_tip.filerev()
+            filetip = self._changectx()[filename]
         except mercurial.revlog.LookupError:
-            add([repo_file])
-            current_page_rev = -1
-        if parent is not None and current_page_rev != parent:
-            msg = self.merge_changes(changectx, repo_file, text, user, parent)
-            user = '<wiki>'
-            text = msg.encode('utf-8')
-        self._commit([repo_file], text, user)
+            if parent_rev != -1:
+                raise IndexError("no such parent revision %r" % parent_rev)
+            return ('tip', None)
+        last_rev = filetip.filerev()
+        if parent_rev > last_rev:
+            raise IndexError("no such parent revision %r" % parent_rev)
+        if parent_rev == last_rev:
+            return ('tip', None)
+        return parent_rev, last_rev
 
-    def _commit(self, files, text, user):
-        try:
-            return self.repo.commit(files=files, text=text, user=user,
-                                    force=True, empty_ok=True)
-        except TypeError:
-            # Mercurial 1.3 doesn't accept empty_ok or files parameter
-            match = mercurial.match.exact(self.repo_path, '', list(files))
-            return self.repo.commit(match=match, text=text, user=user,
-                                    force=True)
+    def _merge(self, repo_file, parent, other, data):
+        filetip = self._changectx()[repo_file]
+        parent_data = filetip.filectx(parent).data()
+        other_data = filetip.filectx(other).data()
+        return merge_func(parent_data, other_data, data)
 
-    def save_data(self, title, data, author=u'', comment=u'', parent=None):
-        """Save data as specified page."""
+    def save_data(self, title, data, author=None, comment=None, parent_rev=None):
+        """Save a new revision of the page. If the data is None, deletes it."""
 
-        try:
-            temp_path = tempfile.mkdtemp(dir=self.path)
-            file_path = os.path.join(temp_path, 'saved')
-            f = open(file_path, "wb")
-            f.write(data)
-            f.close()
-            self.save_file(title, file_path, author, comment, parent)
-        finally:
-            try:
-                os.unlink(file_path)
-            except OSError:
-                pass
-            try:
-                os.rmdir(temp_path)
-            except OSError:
-                pass
+        _ = self._
+        user = (author or _(u'anon')).encode('utf-8')
+        text = (comment or _(u'comment')).encode('utf-8')
+        repo_file = self._title_to_file(title)
+        parent, other = self._get_parents(repo_file, parent_rev)
+        if data is None:
+            if title not in self:
+                raise error.ForbiddenErr()
+        else:
+            if other is not None:
+                try:
+                    data = self._merge(repo_file, parent, other, data)
+                except ValueError:
+                    text = _(u'failed merge of edit conflict').encode('utf-8')
+        def filectxfn(repo, memctx, path):
+            if data is None:
+                raise IOError()
+            return mercurial.context.memfilectx(path, data, False, False, None)
+        ctx = mercurial.context.memctx(
+            repo=self.repo,
+            parents=(parent, other),
+            text=text,
+            files=[repo_file],
+            filectxfn=filectxfn,
+            user=user,
+        )
+        self.repo.commitctx(ctx)
+        self._tips = {}
+
+    def delete_page(self, title, author, comment):
+        self.save_data(title, None, author, comment)
 
     def save_text(self, title, text, author=u'', comment=u'', parent=None):
         """Save text as specified page, encoded to charset."""
@@ -294,57 +239,21 @@ class WikiStorage(object):
     def page_text(self, title):
         """Read unicode text of a page."""
 
-        data = self.open_page(title).read()
+        data = self.page_data(title)
         text = unicode(data, self.charset, 'replace')
         return text
 
-    def page_lines(self, page):
-        for data in page.xreadlines():
-            yield unicode(data, self.charset, 'replace')
-
-    @locked_repo
-    def delete_page(self, title, author=u'', comment=u''):
-        user = author.encode('utf-8') or 'anon'
-        text = comment.encode('utf-8') or 'deleted'
-        repo_file = self._title_to_file(title)
-        file_path = self._file_path(title)
-        self._check_path(file_path)
-        try:
-            # Mercurial 1.5 and earlier have .remove() on the repo
-            remove = self.repo.remove
-        except AttributeError:
-            # Mercurial 1.6 and later
-            try:
-                remove = self.repo[None].remove
-            except AttributeError:
-                # Mercurial 1.9 and later
-                remove = self.repo[None].forget
-        try:
-            os.unlink(file_path)
-        except OSError:
-            pass
-        remove([repo_file])
-        self._commit([repo_file], text, user)
-
     def open_page(self, title):
         """Open the page and return a file-like object with its contents."""
+        return StringIO.StringIO(self.page_data(title))
 
-        file_path = self._file_path(title)
-        self._check_path(file_path)
+    def page_data(self, title):
+        repo_file = self._title_to_file(title)
         try:
-            return open(file_path, "rb")
-        except IOError:
+            filetip = self._changectx()[repo_file]
+        except mercurial.revlog.LookupError:
             raise error.NotFoundErr()
-
-    def page_file_meta(self, title):
-        """Get page's inode number, size and last modification time."""
-
-        try:
-            (st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
-             st_atime, st_mtime, st_ctime) = os.stat(self._file_path(title))
-        except OSError:
-            return 0, 0, 0
-        return st_ino, st_size, st_mtime
+        return filetip.data()
 
     def page_meta(self, title):
         """Get page's revision, date, last editor and his edit comment."""
@@ -369,12 +278,18 @@ class WikiStorage(object):
     def _changectx(self):
         """Get the changectx of the tip."""
 
+        thread_id = thread.get_ident()
         try:
-            # This is for Mercurial 1.0
-            return self.repo.changectx()
-        except TypeError:
-            # Mercurial 1.3 (and possibly earlier) needs an argument
-            return self.repo.changectx('tip')
+            return self._tips[thread_id]
+        except KeyError:
+            try:
+                # This is for Mercurial 1.0
+                tip = self.repo.changectx()
+            except TypeError:
+                # Mercurial 1.3 (and possibly earlier) needs an argument
+                tip = self.repo.changectx('tip')
+            self._tips[thread_id] = tip
+            return tip
 
     def _find_filectx(self, title):
         """Find the last revision in which the file existed."""
@@ -416,7 +331,7 @@ class WikiStorage(object):
             raise error.NotFoundErr()
         try:
             data = filectx_tip.filectx(rev).data()
-        except IndexError:
+        except LookupError:
             raise error.NotFoundErr()
         return data
 
@@ -451,13 +366,12 @@ class WikiStorage(object):
     def all_pages(self):
         """Iterate over the titles of all pages in the wiki."""
 
-        for filename in os.listdir(self.path):
-            file_path = os.path.join(self.path, filename)
-            file_repopath = os.path.join(self.repo_prefix, filename)
-            if (os.path.isfile(file_path)
-                and not os.path.islink(file_path)
-                and not filename.startswith('.')):
-                yield self._file_to_title(file_repopath)
+        for repo_file in self._changectx():
+            if (repo_file.startswith(self.repo_prefix) and
+                '/' not in repo_file[len(self.repo_prefix):]):
+                title = self._file_to_title(repo_file)
+                if title in self:
+                    yield title
 
     def changed_since(self, rev):
         """
@@ -491,6 +405,13 @@ class WikiSubdirectoryStorage(WikiStorage):
     # TODO: make them configurable
     index = "Index"
 
+    def _is_directory(self, repo_path):
+        """Checks whether the path is a directory in the repository."""
+
+        if not repo_path:
+            return True
+        return repo_path in self._changectx().dirs()
+
     def _title_to_file(self, title):
         """
         Modified escaping allowing (some) slashes and spaces.
@@ -502,7 +423,7 @@ class WikiSubdirectoryStorage(WikiStorage):
         escaped = self.periods_re.sub('%2E', escaped)
         escaped = self.slashes_re.sub('%2F', escaped)
         path = os.path.join(self.repo_prefix, escaped)
-        if os.path.isdir(os.path.join(self.repo_path, path)):
+        if self._is_directory(path):
             path = os.path.join(path, self.index)
         if page.page_mime(title) == 'text/x-wiki' and self.extension:
             path += self.extension
@@ -515,88 +436,60 @@ class WikiSubdirectoryStorage(WikiStorage):
             filepath = os.path.dirname(filepath)
         return super(WikiSubdirectoryStorage, self)._file_to_title(filepath)
 
-    def turn_into_subdirectory(self, path):
-        """Turn a single-file page into an index page inside a subdirectory."""
-
-        _ = self._
-        self._check_path(path)
-        dir_path = os.path.dirname(path)
-        if not os.path.isdir(dir_path):
-            self.turn_into_subdirectory(dir_path)
-        if not os.path.exists(path):
-            os.mkdir(path)
-            return
-        try:
-            temp_dir = tempfile.mkdtemp(dir=self.path)
-            temp_path = os.path.join(temp_dir, 'saved')
-            mercurial.commands.rename(self.ui, self.repo, path, temp_path)
-            os.makedirs(path)
-            index_path = os.path.join(path, self.index)
-            mercurial.commands.rename(self.ui, self.repo, temp_path,
-                                      index_path)
-        finally:
-            try:
-                os.rmdir(temp_dir)
-            except OSError:
-                pass
-        def repo_path(path):
-            return path[len(self.repo_path)+1:]
-        files = [repo_path(index_path), repo_path(path)]
-        self._commit(files, _(u"made subdirectory page"), "<wiki>")
-
-    @locked_repo
-    def save_file(self, title, file_name, author=u'', comment=u'',
-                  parent=None):
+    def save_data(self, title, data, author=u'', comment=u'', parent_rev=None):
         """Save the file and make the subdirectories if needed."""
 
-        path = self._file_path(title)
-        dir_path = os.path.dirname(path)
-        if not os.path.isdir(dir_path):
-            self.turn_into_subdirectory(dir_path)
-        try:
-            os.makedirs(os.path.join(self.repo_path, dir_path))
-        except OSError, e:
-            if e.errno != errno.EEXIST:
-                # "File exists"
-                raise
-        super(WikiSubdirectoryStorage, self).save_file(title, file_name,
-                                                       author, comment, parent)
-
-    @locked_repo
-    def delete_page(self, title, author=u'', comment=u''):
-        """
-        Remove empty directories after deleting a page.
-
-        Note that Mercurial doesn't track directories, so we don't have to
-        commit after removing empty directories.
-        """
-
-        super(WikiSubdirectoryStorage, self).delete_page(title, author,
-                                                         comment)
-        file_path = self._file_path(title)
-        self._check_path(file_path)
-        dir_path = os.path.dirname(file_path)
-        if dir_path != self.repo_path:
-            try:
-                os.removedirs(dir_path)
-            except OSError, e:
-                if e.errno != errno.ENOTEMPTY:
-                    # "Directory not empty"
-                    raise
+        _ = self._
+        user = (author or _(u'anon')).encode('utf-8')
+        text = (comment or _(u'comment')).encode('utf-8')
+        repo_file = self._title_to_file(title)
+        files = [repo_file]
+        dir_path = None
+        new_dir_path = None
+        if os.path.basename(repo_file) != self.index:
+            # Move a colliding file out of the way.
+            dir_path = os.path.dirname(repo_file)
+            while dir_path:
+                if dir_path in self._changectx():
+                    new_dir_path = os.path.join(dir_path, self.index)
+                    files.extend([dir_path, new_dir_path])
+                    dir_data = self._changectx()[dir_path].data()
+                    break
+                dir_path = os.path.dirname(dir_path)
+        parent, other = self._get_parents(repo_file, parent_rev)
+        if data is None:
+            if title not in self:
+                raise error.NotFoundErr()
+        else:
+            if other is not None:
+                try:
+                    data = self._merge(repo_file, parent, other, data)
+                except ValueError:
+                    text = _(u'failed merge of edit conflict').encode('utf-8')
+        def filectxfn(repo, memctx, path):
+            if data is None or path == dir_path:
+                raise IOError()
+            if path == new_dir_path:
+                return mercurial.context.memfilectx(path, dir_data,
+                        False, False, dir_path)
+            return mercurial.context.memfilectx(path, data, False, False, None)
+        ctx = mercurial.context.memctx(
+            repo=self.repo,
+            parents=(parent, other),
+            text=text,
+            files=files,
+            filectxfn=filectxfn,
+            user=user,
+        )
+        self.repo.commitctx(ctx)
+        self._tips = {}
 
     def all_pages(self):
-        """
-        Iterate over the titles of all pages in the wiki.
-        Include subdirectories but skip over index.
-        """
+        """Iterate over the titles of all pages in the wiki."""
+
+        for repo_file in self._changectx():
+            if repo_file.startswith(self.repo_prefix):
+                title = self._file_to_title(repo_file)
+                if title in self:
+                    yield title
 
-        for (dirpath, dirnames, filenames) in os.walk(self.path):
-            path = dirpath[len(self.path) + 1:]
-            if path.startswith('.'):
-                continue
-            for name in filenames:
-                filepath = os.path.join(dirpath, name)
-                repopath = os.path.join(self.repo_prefix, path, name)
-                if (os.path.isfile(filepath)
-                    and not name.startswith('.')):
-                    yield self._file_to_title(repopath)
